//
//  XWGrid.m
//  Crossword
//
//  Created by Max Williams on 28/12/2008.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

#import "XWGrid.h"


@implementation XWGrid

@synthesize width, height, cells;

- (id)initWithWidth:(unsigned int)initWidth height:(unsigned int)initHeight {
	if (![super init]) return nil;
	width = initWidth;
	height = initHeight;
	NSMutableArray * tempCells = [NSMutableArray arrayWithCapacity:width*height];
	for (int i = 0; i<width*height; i++) {
		[tempCells addObject:[XWCell blankCell]];
	}
	cells = [NSArray arrayWithArray:tempCells];
	[cells retain];
	
	undoManager = [[NSUndoManager alloc] init];
	[undoManager prepareWithInvocationTarget:self];
	
	return self;
}

- (void)dealloc {
	[cells release];
	[undoManager release];
	[super dealloc];
}

- (void)undoLastWord {
	if ([undoManager canUndo]) {
		[undoManager undo];
		return;
	}
	else {
		NSLog(@"undo problem");
	}
}

- (NSString *)description {
	NSMutableString * desc = [NSMutableString stringWithCapacity:(width+2)*height*4];
	XWCell * cell;
	for (int y = 0; y<height; y++) {
		[desc appendString:@"\n"];
		for (int x=0; x<width; x++) {
			cell = [cells objectAtIndex:(width*y+x)];
			if ([cell type]==XWLetterCell)
				[desc appendFormat:@" %@",[self letterAtX:x Y:y]];
			else if ([cell type]==XWBlankCell)
				[desc appendString:@" ·"];
			else if ([cell type]==XWBlockingCell)
				[desc appendString:@" •"];
		}
	}
	return desc;
}

- (XWCell *)cellAtX:(int)x Y:(int)y {	
	if (x>=width) return nil;
	if (y>=height) return nil;
	return [cells objectAtIndex:(width*y+x)];
}

- (XWLetter *)letterAtX:(int)x Y:(int)y {	
	XWCell * cell = [self cellAtX:x Y:y];
	if (cell == nil) return nil;
	if ([cell type]!=XWLetterCell) return nil;
	return [cell letter];
}

- (void)setLetter:(XWLetter *)letter atX:(int)x Y:(int)y {
	XWCell * cell = [self cellAtX:x Y:y];
	cell.letter = letter;
}

- (BOOL)word:(XWWord *)word fitsAt:(XWLocation)loc {
	XWLocation * letterLocations = calloc(sizeof(XWLocation),word.length);
	for (int i = 0; i<word.length; i++) {
		letterLocations[i].x = loc.x;
		letterLocations[i].y = loc.y;
		if (loc.orientation == XWOrientationHorizontal) letterLocations[i].x += i;
		if (loc.orientation == XWOrientationVertical) letterLocations[i].y += i;
	}
	
	// check if fits inside grid
	
	if (loc.x<0 || loc.y<0) return NO;
	
	if (loc.orientation == XWOrientationHorizontal) {
		if (loc.x+word.length > width) return NO;
	}
	if (loc.orientation == XWOrientationVertical) {
		if (loc.y+word.length > height) return NO;
	}
	
	// the first piece can go anywhere
	
	BOOL otherLetters = NO;
	
	for (XWCell * cell in self.cells) {
		if (cell.type == XWLetterCell) otherLetters = YES;
	}
	
	if (otherLetters == NO) return YES;
	
	// otherwise the word must intersect with at least one letter
	
	int intersectCount = 0;
	
	for (int i = 0; i<word.length; i++) {
		XWLetter * testLetter = [self letterAtX:letterLocations[i].x Y:letterLocations[i].y];
		if (testLetter != nil) {
			intersectCount++;
		}
	}
	
	if (intersectCount == 0) return NO;
	
	// cells at either end of new word must be empty
	
	if (loc.orientation == XWOrientationHorizontal) {
		if ([self letterAtX:loc.x-1 Y:loc.y] != nil) return NO;
		if ([self letterAtX:loc.x+word.length Y:loc.y] != nil) return NO;
	}
	if (loc.orientation == XWOrientationVertical) {
		if ([self letterAtX:loc.x Y:loc.y-1] != nil) return NO;
		if ([self letterAtX:loc.x Y:loc.y+word.length] != nil) return NO;
	}
	
	// check for incorrect overlapping letters
	// also take this opportunity to work out exempt cells
	
	NSMutableArray * exemptCells = [NSMutableArray arrayWithCapacity:10];
	
	for (int i = 0; i<word.length; i++) {
		XWLetter * testLetter = [self letterAtX:letterLocations[i].x Y:letterLocations[i].y];
		if (testLetter != nil) {
			if (testLetter != [word letterAtIndex:i]) return NO;
			// correct overlap; two adjacent cells are exempt from follwing test
			if (loc.orientation == XWOrientationHorizontal) {
				XWCell * cellAbove = [self cellAtX:letterLocations[i].x Y:letterLocations[i].y-1];
				XWCell * cellBelow = [self cellAtX:letterLocations[i].x Y:letterLocations[i].y+1];
				if (cellAbove != nil) [exemptCells addObject:cellAbove];
				if (cellBelow != nil) [exemptCells addObject:cellBelow];
			}
			if (loc.orientation == XWOrientationVertical) {
				XWCell * cellLeft = [self cellAtX:letterLocations[i].x-1 Y:letterLocations[i].y];
				XWCell * cellRight = [self cellAtX:letterLocations[i].x+1 Y:letterLocations[i].y];
				if (cellLeft != nil) [exemptCells addObject:cellLeft];
				if (cellRight != nil) [exemptCells addObject:cellRight];
			}
			// cannot have two overlapping cells next to each other
			if (loc.orientation == XWOrientationHorizontal) {
				XWLetter * cellLeft = [self letterAtX:letterLocations[i].x-1 Y:letterLocations[i].y];
				XWLetter * cellRight = [self letterAtX:letterLocations[i].x+1 Y:letterLocations[i].y];
				if (cellLeft != nil) return NO;
				if (cellRight != nil) return NO;
			}
			if (loc.orientation == XWOrientationVertical) {
				XWLetter * cellAbove = [self letterAtX:letterLocations[i].x Y:letterLocations[i].y-1];
				XWLetter * cellBelow = [self letterAtX:letterLocations[i].x Y:letterLocations[i].y+1];
				if (cellAbove != nil) return NO;
				if (cellBelow != nil) return NO;
			}
			
		}
	}
	
	// check adjacent rows/columns
	
	
	for (int i = 0; i<word.length; i++) {
		XWLocation testLoc1 = loc;
		XWLocation testLoc2 = loc;
		
		if (loc.orientation == XWOrientationHorizontal) {
			testLoc1.x += i;
			testLoc2.x += i;
			testLoc1.y -= 1;
			testLoc2.y += 1;
		}
		if (loc.orientation == XWOrientationVertical) {
			testLoc1.y += i;
			testLoc2.y += i;
			testLoc1.x -= 1;
			testLoc2.x += 1;
		}
		
		
		if ([self letterAtX:testLoc1.x Y:testLoc1.y]!=nil) {
			if ([exemptCells containsObject:[self cellAtX:testLoc1.x Y:testLoc1.y]] == NO) {
				return NO;
			}
		}
		if ([self letterAtX:testLoc2.x Y:testLoc2.y]!=nil) {
			if ([exemptCells containsObject:[self cellAtX:testLoc2.x Y:testLoc2.y]] == NO) {
				return NO;
			}
		}
	}
	
	free(letterLocations);
	return YES;
}

- (void)setBlankCellAtX:(int)x Y:(int)y {
	[self cellAtX:x Y:y].letter = nil;
	[self cellAtX:x Y:y].type = XWBlankCell;
}

- (void)addCellsFromWord:(XWWord *)word atLocation:(XWLocation)loc {
	[undoManager beginUndoGrouping];
	for (int letterIndex=0; letterIndex<word.length; letterIndex++) {
		XWLetter * letter = [word letterAtIndex:letterIndex];
		int x = loc.x;
		int y = loc.y;
		if (loc.orientation == XWOrientationHorizontal)
			x += letterIndex;
		if (loc.orientation == XWOrientationVertical)
			y += letterIndex;
		if ([self letterAtX:x Y:y] == nil) {
			[self setLetter:letter atX:x Y:y];
			[[undoManager prepareWithInvocationTarget:self] setBlankCellAtX:x Y:y];
		}
	}
	[undoManager endUndoGrouping];
}


@end
